@@ -1,26 +1,34 @@
-package Pcore::WebDriver v0.1.4;
+package Pcore::WebDriver v0.2.0;
 
-use Pcore -dist, -const, -class, -result,
+use Pcore -dist, -const, -role, -result,
   -export => {
-    WD_DRIVER  => [qw[$WD_PHANTOMJS $WD_CHROME $WD_FIREFOX]],
     WD_LOCATOR => [qw[$WD_CLASS_NAME $WD_CSS_SELECTOR $WD_ID $WD_NAME $WD_LINK_TEXT $WD_LINK_TEXT_PART $WD_TAG_NAME $WD_XPATH]],
     WD_KEY     => [qw[$WD_KEY]],
-    CONST      => [qw[:WD_DRIVER :WD_LOCATOR :WD_KEY]],
+    CONST      => [qw[:WD_LOCATOR :WD_KEY]],
   };
 use Pcore::Util::Data qw[to_json from_json from_b64];
 require Pcore::WebDriver::WebElement;
 
-const our $WD_PHANTOMJS => 1;
-const our $WD_CHROME    => 2;
-const our $WD_FIREFOX   => 3;
+with qw[Pcore::Util::Result::Status];
 
-has driver => ( is => 'ro', isa => Enum [ $WD_PHANTOMJS, $WD_CHROME, $WD_FIREFOX ], required => 1 );
+requires qw[get_cmd get_desired_capabilities];
 
-has host => ( is => 'ro', isa => Str, default => '127.0.0.1' );
-has port => ( is => 'ro', isa => PositiveInt );
+has host => ( is => 'ro', isa => Str,         required => 1 );
+has port => ( is => 'ro', isa => PositiveInt, required => 1 );
+
+has desired_capabilities => ( is => 'ro', isa => Maybe [HashRef] );
+
+has disable_javascript   => ( is => 'ro', isa => Bool, default => 0 );
+has disable_images       => ( is => 'ro', isa => Bool, default => 0 );
+has enable_flash         => ( is => 'ro', isa => Bool, default => 0 );
+has enable_notifications => ( is => 'ro', isa => Bool, default => 0 );
+has useragent            => ( is => 'ro', isa => Str );
+
+has proxy_type => ( is => 'ro', isa => Enum [qw[direct manual]], default => 'direct' );
+has http_proxy => ( is => 'ro', isa => Str );    # '192.168.175.1:9070'
+has ssl_proxy  => ( is => 'ro', isa => Str );    # '192.168.175.1:9070'
 
 has session_id => ( is => 'ro', isa => Str, init_arg => undef );
-
 has _proc => ( is => 'ro', isa => InstanceOf ['Pcore::Util:PM::Proc'], init_arg => undef );    # proc handle
 
 const our $WD_STATUS_REASON => {
@@ -258,59 +266,32 @@ sub update_all ( $self, $cb = undef ) {
     return $blocking_cv ? $blocking_cv->recv : ();
 }
 
-sub new_phantomjs ( $self, @args ) {
-    return $self->new( $WD_PHANTOMJS, @args );
-}
+sub DEMOLISH ( $self, $global ) {
+    if ( !$global && $self->{session_id} ) {
+        P->http->_request(
+            method  => 'DELETE',
+            url     => "http://$self->{host}:$self->{port}/session/$self->{session_id}",
+            timeout => 0,
+        );
+    }
 
-sub new_chrome ( $self, @args ) {
-    return $self->new( $WD_CHROME, @args );
-}
-
-sub new_firefox ( $self, @args ) {
-    return $self->new( $WD_FIREFOX, @args );
+    return;
 }
 
 # TODO sleep
-around new => sub ( $orig, $self, $driver, %args ) {
+around new => sub ( $orig, $self, $args ) {
     my $blocking_cv = defined wantarray ? AE::cv : undef;
 
-    $self = $self->$orig(
-        {   driver => $driver,
-            host   => $args{host} || '127.0.0.1',
-        }
-    );
+    $args->{host} ||= '127.0.0.1';
 
-    $self->{port} = $args{port} || ( P->sys->get_free_port( $self->{host} ) // die q[Error get free port] );
+    $args->{port} ||= ( P->sys->get_free_port( $args->{host} ) // die q[Error get free port] );
 
-    my $driver_class;
+    $args->{status} = 200;
 
-    if ( $self->is_phantomjs ) {
-        require Pcore::WebDriver::Driver::PhantomJS;
-
-        $driver_class = 'Pcore::WebDriver::Driver::PhantomJS';
-    }
-    elsif ( $self->is_chrome ) {
-        require Pcore::WebDriver::Driver::Chrome;
-
-        $driver_class = 'Pcore::WebDriver::Driver::Chrome';
-    }
-    elsif ( $self->is_firefox ) {
-        require Pcore::WebDriver::Driver::Firefox;
-
-        $driver_class = 'Pcore::WebDriver::Driver::Firefox';
-    }
-    else {
-        die q[Unknown web driver];
-    }
-
-    $driver = $driver_class->new(
-        {   host => $self->{host},
-            port => $self->{port},
-        }
-    );
+    $self = $self->$orig($args);
 
     P->pm->run_proc(
-        $driver->cmd,
+        $self->get_cmd,
         win32_create_no_window => 1,
         on_ready               => sub ($proc) {
             $self->{_proc} = $proc;
@@ -319,10 +300,11 @@ around new => sub ( $orig, $self, $driver, %args ) {
             sleep 1;
 
             $self->new_session(
-                sub ($res) {
-                    $args{on_ready}->($self) if $args{on_ready};
+                undef,
+                sub ($session) {
+                    $args->{on_ready}->($session) if $args->{on_ready};
 
-                    $blocking_cv->($self) if $blocking_cv;
+                    $blocking_cv->($session) if $blocking_cv;
 
                     return;
                 }
@@ -335,16 +317,35 @@ around new => sub ( $orig, $self, $driver, %args ) {
     return $blocking_cv ? $blocking_cv->recv : ();
 };
 
-sub is_phantomjs ($self) {
-    return $self->{driver} == $WD_PHANTOMJS;
+around get_desired_capabilities => sub ( $orig, $self, $desired_capabilities = undef ) {
+    my $cap = P->hash->merge(
+        $self->{desired_capabilities} // {},
+        $desired_capabilities // {},
+        {   javascriptEnabled => $self->disable_javascript ? \0 : \1,
+            acceptSslCerts    => \1,
+            takesScreenshot   => \1,
+            handlesAlerts     => \1,
+            proxy             => {
+                proxyType => $self->proxy_type,
+                httpProxy => $self->http_proxy,
+                sslProxy  => $self->ssl_proxy,
+            },
+        },
+    );
+
+    return $self->$orig($cap);
+};
+
+sub new_phantomjs ( $self, %args ) {
+    return P->class->load('Pcore::WebDriver::PhantomJS')->new( \%args );
 }
 
-sub is_chrome ($self) {
-    return $self->{driver} == $WD_CHROME;
+sub new_chrome ( $self, %args ) {
+    return P->class->load('Pcore::WebDriver::Chrome')->new( \%args );
 }
 
-sub is_firefox ($self) {
-    return $self->{driver} == $WD_FIREFOX;
+sub new_firefox ( $self, %args ) {
+    return P->class->load('Pcore::WebDriver::Firefox')->new( \%args );
 }
 
 sub _send_command ( $self, $method, $path, $body, $cb = undef, $processing_cb = undef ) {
@@ -367,7 +368,7 @@ sub _send_command ( $self, $method, $path, $body, $cb = undef, $processing_cb =
                 $res = result [ $http_res->status, $http_res->reason ];
             }
 
-            $res->{session_id} = $self->{session_id} // $body->{sessionId};
+            $res->{session_id} = $body->{sessionId};
 
             $res->{data} = $body->{value};
 
@@ -397,45 +398,33 @@ sub _send_command ( $self, $method, $path, $body, $cb = undef, $processing_cb =
 }
 
 # SESSIONS
-sub new_session ( $self, $cb = undef ) {
+sub new_session ( $self, $desired_capabilities = undef, $cb = undef ) {
     return $self->_send_command(
         'POST',
         '/session',
-        {   desiredCapabilities => {
-                javascriptEnabled => \1,
-                acceptSslCerts    => \1,
-                takesScreenshot   => \1,
-                handlesAlerts     => \1,
-                chromeOptions     => {
-                    args  => ['start-maximized'],    # 'no-sandbox', 'user-data-dir'
-                    prefs => {                       # list of options can be found in Preferences file in chrome profile
-                        'profile.default_content_setting_values.notifications' => 2,
-                    },
-                },
-                'moz:firefoxOptions' => {
-                    binary => '',
-                    args   => ['--devtools'],
-                },
-            },
-        },
+        { desiredCapabilities => $self->get_desired_capabilities($desired_capabilities) },
         $cb,
         sub ( $res, $cb ) {
-            $self->{session_id} = $res->{session_id} if $res;
+            my $session;
 
-            $cb->($res);
+            if ( $self->{session_id} ) {
+                $session = bless { $self->%* }, ref $self;
+            }
+            else {
+                $session = $self;
+            }
+
+            $session->set_status( [ $res->status, $res->reason ] );
+
+            $session->{session_id} = $res->{session_id};
+
+            $cb->($session);
 
             return;
         }
     );
 }
 
-# TODO
-sub delete_session ( $self, $cb = undef ) {
-    ...;
-
-    return;
-}
-
 sub status ( $self, $cb = undef ) {
     return $self->_send_command( 'GET', '/status', undef, $cb );
 }
@@ -1028,17 +1017,15 @@ JS
 ## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
 ## | Sev. | Lines                | Policy                                                                                                         |
 ## |======+======================+================================================================================================================|
-## |    3 | 153, 192, 231        | ErrorHandling::RequireCheckingReturnValueOfEval - Return value of eval not tested                              |
+## |    3 | 161, 200, 239        | ErrorHandling::RequireCheckingReturnValueOfEval - Return value of eval not tested                              |
 ## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
-## |    3 | 350                  | Subroutines::ProhibitManyArgs - Too many arguments                                                             |
+## |    3 | 271, 354             | Subroutines::ProtectPrivateSubs - Private subroutine/method used                                               |
 ## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
-## |    3 | 353                  | Subroutines::ProtectPrivateSubs - Private subroutine/method used                                               |
+## |    3 | 351                  | Subroutines::ProhibitManyArgs - Too many arguments                                                             |
 ## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
-## |    3 | 434, 713, 719, 725,  | ControlStructures::ProhibitYadaOperator - yada operator (...) used                                             |
-## |      | 731, 737, 746, 752,  |                                                                                                                |
-## |      | 758, 764             |                                                                                                                |
-## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
-## |    2 | 416                  | ValuesAndExpressions::ProhibitEmptyQuotes - Quotes used with a string containing no non-whitespace characters  |
+## |    3 | 702, 708, 714, 720,  | ControlStructures::ProhibitYadaOperator - yada operator (...) used                                             |
+## |      | 726, 735, 741, 747,  |                                                                                                                |
+## |      | 753                  |                                                                                                                |
 ## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
 ##
 ## -----SOURCE FILTER LOG END-----
